/*
 * libinit.c - initialization code for the shared library
 *
 * Copyright (C) 2018 Thorsten Otto
 */

#include "tiffiop.h"
#include <stdio.h>
#include <mint/basepage.h>
#include <mint/mintbind.h>
#include <mint/slb.h>
#include <mint/cookie.h>
#include <errno.h>
#include <slb/zlib.h>
#include <slb/jpeg.h>
#include <slb/lzma.h>
#include "tiff.h"
#include "slbload.h"
#include "libstrct.h"
#include "vernum.h"

long slb_init(void);
void slb_exit(void);
long slb_open(BASEPAGE *bp);
void slb_close(BASEPAGE *bp);
extern char const slb_header[];
static const BASEPAGE *my_base;


#ifdef __MSHORT__
# error "the tiff.slb must not be compiled with -mshort"
#endif

#define MINT_MAX_PID 999
#define MAX_PIDS (MINT_MAX_PID + 1)

struct per_proc {
	struct _tiff_funcs *funcs;
	pid_t pid;
};

static struct per_proc procs[MAX_PIDS];

static pid_t slb_user(void)
{
	pid_t pid = Pgetpid();
	if (pid == (pid_t)-ENOSYS)
		pid = 1;
	return pid;
}


static struct per_proc *get_proc(pid_t pid, pid_t slot)
{
	int i;
	
	/*
	 * see if we can use it as a direct index into our array
	 */
	if (pid > 0 && pid < MAX_PIDS && procs[pid].pid == slot)
		return &procs[pid];
	/*
	 * TODO: if MiNT ever uses pids >= 1000,
	 * use a hash instead
	 */
	for (i = 1; i < MAX_PIDS; i++)
		if (procs[i].pid == slot)
			return &procs[i];
	return NULL;
}


__attribute__((__noinline__))
struct _tiff_funcs *tiff_get_slb_funcs(void)
{
	pid_t pid = slb_user();
	struct per_proc *proc = get_proc(pid, pid);
	if (proc == NULL)
		return NULL;
	return proc->funcs;
}


/*
 * these are not optional and cannot be set
 * to zero in the header, even if they
 * currently don't do anything
 */
long slb_init(void)
{
	const BASEPAGE *bp;
	const long *exec_longs;

	bp = (BASEPAGE *)(slb_header - 256);
	exec_longs = (const long *)((const char *)bp + 28);
	if ((exec_longs[0] == 0x283a001aL && exec_longs[1] == 0x4efb48faL) ||
		(exec_longs[0] == 0x203a001aL && exec_longs[1] == 0x4efb08faL))
		bp = (const BASEPAGE *)((const char *)bp - 228);
	my_base = bp;
	return 0;
}


void slb_exit(void)
{
}


long slb_open(BASEPAGE *bp)
{
	pid_t pid = slb_user();
	struct per_proc *proc = get_proc(pid, 0);
	
	(void)(bp);
	if (proc == NULL)
		return -ENFILE; /* ENFILE: file table overflow, well, proc table in this case */
	/*
	 * check if SLB is already in use by this process;
	 * should not happen since MiNT should have taken care of that already
	 */
	if (proc->funcs != NULL)
		return -EACCES;
	
	proc->pid = pid;
	
	return 0;
}


void slb_close(BASEPAGE *bp)
{
	pid_t pid = slb_user();
	struct per_proc *proc = get_proc(pid, pid);

	(void)(bp);
	if (proc == NULL)
		return;
	proc->funcs = NULL;
	proc->pid = 0;
}


/*
 * get the function table pointer passed from the application.
 * Automatically done in slb_zlib_open()
 */
void unixErrorHandler(const char* module, const char* fmt, va_list ap);
void unixWarningHandler(const char* module, const char* fmt, va_list ap);

__attribute__((__noinline__))
static long tiff_set_imports(struct _tiff_funcs *funcs)
{
	pid_t pid = slb_user();
	struct per_proc *proc = get_proc(pid, pid);

	if (proc == NULL)
		return -ESRCH;
	if (funcs->struct_size != sizeof(*funcs))
		return -EINVAL;
	if (funcs->tiff_vernum > TIFF_VERSION)
		return -EBADARG;
	if (funcs->int_size != sizeof(int))
		return -ERANGE;
	proc->funcs = funcs;
	funcs->_TIFFwarningHandler = unixWarningHandler;
	funcs->_TIFFwarningHandlerExt = 0;
	funcs->_TIFFerrorHandler = unixErrorHandler;
	funcs->_TIFFerrorHandlerExt = 0;
	funcs->_TIFFextender = 0;
	return 0;
}
		

__attribute__((__noinline__))
static long slb_compile_flags(void)
{
	long flags = 0;
#if defined(__mc68020__) || defined(__mc68030__) || defined(__mc68040__) || defined(__mc68060__) || defined(__mc68080__) || defined(__apollo__)
    flags |= (1L << 16);
#endif
#if defined(__mcoldfire__)
	/*
	 * use a large value here, so that we can call this function also on
	 * non-coldfire hardware; the compiler otherwise would generate a mov3q instruction
	 */
    flags |= (1L << 17);
#endif
	return flags;
}


long __CDECL tiff_slb_control(long fn, void *arg)
{
	switch ((int)fn)
	{
	case 0:
		return slb_compile_flags();
	case 1:
		return tiff_set_imports(arg);
	case 2:
		return (long)my_base;
	case 3:
		return (long)slb_header;
	case 4:
		return (long)my_base->p_cmdlin;
	case 5:
		return (long)(
#if (defined(ZIP_SUPPORT) || defined(PIXARLOG_SUPPORT)) && defined(ZLIB_SLB)
			ZLIB_SHAREDLIB_NAME "\0"
#endif
#if defined(LZMA_SUPPORT) && defined(LZMA_SLB)
			LZMA_SHAREDLIB_NAME "\0"
#endif
#if (defined(JPEG_SUPPORT) || defined(OJPEG_SUPPORT)) && defined(JPEG_SLB)
			JPEG_SHAREDLIB_NAME "\0"
#endif
#if defined(ZSTD_SUPPORT) && defined(ZSTD_SLB)
			ZSTD_SHAREDLIB_NAME "\0"
#endif
#if defined(WEBP_SUPPORT) && defined(WEBP_SLB)
			WEBP_SHAREDLIB_NAME "\0"
#endif
			"");
	}
	return -ENOSYS;
}


#if (defined(ZIP_SUPPORT) || defined(PIXARLOG_SUPPORT)) && defined(ZLIB_SLB)
long slb_tiff_zlib_open(void)
{
	struct _tiff_funcs *funcs;
	
	funcs = tiff_get_slb_funcs();
	return funcs->p_slb_tiff_zlib_open();
}


/*
 * override SLB library pointer for zlib
 */
SLB *slb_zlib_get(void)
{
	return tiff_get_slb_funcs()->zlib;
}
#endif


#if (defined(JPEG_SUPPORT) || defined(OJPEG_SUPPORT)) && defined(JPEG_SLB)
long slb_tiff_jpeg_open(void)
{
	struct _tiff_funcs *funcs;
	
	funcs = tiff_get_slb_funcs();
	return funcs->p_slb_tiff_jpeg_open();
}


/*
 * override SLB library pointer for jpeg
 */
SLB *slb_jpeglib_get(void)
{
	return tiff_get_slb_funcs()->jpeg;
}
#endif


#if defined(LZMA_SUPPORT) && defined(LZMA_SLB)
long slb_tiff_lzma_open(void)
{
	struct _tiff_funcs *funcs;
	
	funcs = tiff_get_slb_funcs();
	return funcs->p_slb_tiff_lzma_open();
}


/*
 * override SLB library pointer for lzma
 */
SLB *slb_lzma_get(void)
{
	return tiff_get_slb_funcs()->lzma;
}
#endif


#if defined(ZSTD_SUPPORT) && defined(ZSTD_SLB)
long slb_tiff_zstd_open(void)
{
	struct _tiff_funcs *funcs;
	
	funcs = tiff_get_slb_funcs();
	return funcs->p_slb_tiff_zstd_open();
}


/*
 * override SLB library pointer for zstd
 */
SLB *slb_zstd_get(void)
{
	return tiff_get_slb_funcs()->zstd;
}
#endif


/*
 * just redefining memcpy is not enough;
 * the compiler will also generate references to it
 */
void *(memcpy)(void *dest, const void *src, size_t len)
{
	return tiff_get_slb_funcs()->p_memcpy(dest, src, len);
}

/* same for memset */
void *(memset)(void *dest, int c, size_t len)
{
	return tiff_get_slb_funcs()->p_memset(dest, c, len);
}

#undef errno
/* referenced only by some math functions */
int errno;


#if defined(ZSTD_SUPPORT) && !defined(ZSTD_SLB)
/*
 * resolve references for statically linked zstd
 */
#undef memmove
void *memmove(void *dest, const void *src, size_t len) __attribute__((alias("memcpy")));

#undef free
void free(void *ptr)
{
	tiff_get_slb_funcs()->p_free(ptr);
}

#undef malloc
void *malloc(size_t size)
{
	return tiff_get_slb_funcs()->p_malloc(size);
}

#undef calloc
void *calloc(size_t nmemb, size_t size)
{
	return tiff_get_slb_funcs()->p_calloc(nmemb, size);
}

#endif
