/*
 * impopen.c - loader for library
 *
 * Copyright (C) 2019 Thorsten Otto
 */
#include "slb/tiff.h"
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <mint/cookie.h>
#include <mint/mintbind.h>
#include <math.h>
#include <setjmp.h>
#include <slb/zlib.h>
#include <slb/jpeg.h>
#include <slb/lzma.h>
#include "libstrct.h"
#include "slbload.h"
#include "tiff.h"
#include "vernum.h"



typedef struct
{
	long key;
	long value;
} COOKIE;


static COOKIE *get_cookie_jar(void)
{
	return (COOKIE *) Setexc(360, (void (*)()) -1L);
}


static int get_cookie(long key, long *value)
{
	COOKIE *cookiejar = get_cookie_jar();

	*value = 0;
	if (cookiejar)
	{
		while (cookiejar->key)
		{
			if (cookiejar->key == key)
			{
				*value = cookiejar->value;
				return 1;
			}
			cookiejar++;
		}
	}
	return 0;
}


#if defined(__MSHORT__) || defined(__PUREC__) || defined(__AHCC__)

static void *__CDECL gcc_memset(void *s, tiff_int_t val, size_t len)
{
	return memset(s, (int)val, len);
}

static void *__CDECL gcc_memcpy(void *d, const void *s, size_t len)
{
	return memcpy(d, s, len);
}

static tiff_int_t __CDECL gcc_memcmp(const void *s, const void *d, size_t len)
{
	return memcmp(s, d, len);
}

static void *__CDECL gcc_memchr(const void *s, tiff_int_t c, size_t len)
{
	return memchr(s, (int)c, len);
}

static size_t __CDECL gcc_strlen(const char *s)
{
	return strlen(s);
}

static char *__CDECL gcc_strcpy(char *d, const char *s)
{
	return strcpy(d, s);
}

static char *__CDECL gcc_strcat(char *d, const char *s)
{
	return strcat(d, s);
}

static tiff_int_t __CDECL gcc_strcmp(const char *d, const char *s)
{
	return strcmp(d, s);
}

static tiff_int_t __CDECL gcc_strncmp(const char *d, const char *s, size_t len)
{
	return strncmp(d, s, len);
}

static void *__CDECL gcc_malloc(size_t s)
{
	return malloc(s);
}

static void *__CDECL gcc_calloc(size_t nitems, size_t s)
{
	return calloc(nitems, s);
}

static void *__CDECL gcc_realloc(void *ptr, size_t s)
{
	return realloc(ptr, s);
}

static void __CDECL gcc_free(void *s)
{
	free(s);
}

#if defined(__PUREC__) || defined(__AHCC__)
#define ESRCH	3	    /* No such process            */
#define EINTR	4		/* Interrupted system call    */
#define ENXIO	6		/* No such device or address  */
#define EFAULT	14		/* Bad address                */
#define ENOTBLK	15		/* Bulk device required       */
#define EBUSY	16		/* Resource is busy           */
#define ENOTTY	25		/* Not a terminal             */
#define ETXTBSY	26		/* Text file is busy          */
#define EFBIG	27		/* File is too large          */
#define EMLINK	31		/* Too many links             */
#define EPIPE	32		/* Broken pipe                */
#define EILSEQ  84      /* Illegal byte sequence      */
#endif

static tiff_int_t __CDECL gcc_get_errno(void)
{
#if defined(__GNUC__) || defined(_GNU_SOURCE)
	/* we are just here because of -mshort */
	return errno;
#elif defined(__PUREC__) || defined(__AHCC__) || defined(LATTICE)
	/* must translate error numbers from Pure-C library to what mintlib expects */
	switch (errno)
	{
		case EPERM: return 38;
		case ENOENT: return 33;
		case ESRCH: return 20;
		case EINTR: return 128;
		case EIO: return 90;
		case ENXIO: return 46;
		case EBADF: return 37;
#if defined(__PUREC__) || defined(__AHCC__)
		case EILLSPE: return 25; /* -> EINVAL */
		case EINVMEM: return 40; /* -> EFAULT */
#endif
		case EFAULT: return 40;
		case ENOTBLK: return 23;
		case EBUSY: return 2;
		case ENOMEM: return 39;
		case EACCES: return 36;
		case EEXIST: return 85;
#if defined(__PUREC__) || defined(__AHCC__)
		case EPLFMT: return 66; /* -> ENOEXEC */
#endif
#if defined(LATTICE)
		case ENOEXEC: return 66;
		case EAGAIN: return 326;
		case EXDEV: return 48;
		case ECHILD: return 48;
#endif
		case ENODEV: return 15;
		case ENOTDIR: return 34;
		case EINVAL: return 25;
		case ENFILE: return 50;
		case EMFILE: return 35;
		case ENOTTY: return 87;
		case ETXTBSY: return 70;
		case EFBIG: return 71;
		case ENOSPC: return 91;
		case ESPIPE: return 6;
		case EROFS: return 13;
		case EMLINK: return 82;
		case EPIPE: return 81;
		case EDOM: return 89;
		case ERANGE: return 88;
#if defined(__PUREC__) || defined(__AHCC__)
		case ENMFILE: return 49;
#endif
#if defined(LATTICE)
		case EDEADLK: return 22;
#endif
		case EILSEQ: return 27;
	}
	return 1;
#else
 # error "NYI"
	return 1;
#endif
}

static void __CDECL gcc_set_errno(e)
{
#if defined(__GNUC__) || defined(_GNU_SOURCE)
	/* we are just here because of -mshort */
	errno = e;
#else
	switch (e)
	{
	default:
		errno = EINVAL;
		break;
	}
#endif
}

static char *__CDECL gcc_strerror(tiff_int_t e)
{
#if defined(__GNUC__) || defined(_GNU_SOURCE)
	return strerror((int)e);
#else
	/* FIXME: would need strerror from mintlib */
	static char errorbuf[14];
	sprintf(errorbuf, "error #%d", (int)e);
	return errorbuf;
#endif
}

#define MINTLIB_O_CREAT  0x000200L
#define MINTLIB_O_TRUNC  0x000400L
#define MINTLIB_O_APPEND 0x001000L
#define MINTLIB_O_EXCL   0x000800L

static tiff_int_t __CDECL gcc_open(const char *filename, tiff_int_t access, ...)
{
	va_list args;
	int real_access;
	mode_t mode = 0644;
	
	va_start(args, access);
	if (access & MINTLIB_O_CREAT)
		mode = va_arg(args, tiff_int_t);
	va_end(args);
	real_access = (int)access & O_ACCMODE; /* these are identical */
	if (access & MINTLIB_O_CREAT)
		real_access |= O_CREAT;
	if (access & MINTLIB_O_TRUNC)
		real_access |= O_TRUNC;
	if (access & MINTLIB_O_APPEND)
		real_access |= O_APPEND;
	if (access & MINTLIB_O_EXCL)
		real_access |= O_EXCL;
	/* others are not use by zlib/gz_open */
	return open(filename, real_access, mode);
}

static tiff_int_t __CDECL gcc_close(tiff_int_t fd)
{
	return close((int)fd);
}

static ssize_t __CDECL gcc_read(tiff_int_t fd, void *buf, size_t len)
{
	return read((int)fd, buf, len);
}

static ssize_t __CDECL gcc_write(tiff_int_t fd, const void *buf, size_t len)
{
	return write((int)fd, buf, len);
}

static off_t __CDECL gcc_lseek(tiff_int_t fd, off_t off, tiff_int_t whence)
{
	return lseek((int)fd, off, (int)whence);
}

static size_t __CDECL gcc_fwrite(const void *buf, size_t size, size_t count, FILE *fp)
{
	return fwrite(buf, size, count, fp);
}

static tiff_int_t __CDECL gcc_sprintf(char *buf, const char *format, ...)
{
	va_list args;
	int ret;
	va_start(args, fmt);
	ret = vsprintf(buf, format, args);
	va_end(args);
	return ret;
}

static tiff_int_t __CDECL gcc_snprintf(char *buf, size_t maxlen, const char *format, ...)
{
	va_list args;
	int ret;
	(void)maxlen;
	va_start(args, fmt);
	ret = vsprintf(buf, format, args);
	va_end(args);
	return ret;
}

static tiff_int_t __CDECL gcc_vsnprintf(char *buf, size_t len, const char *format, va_list args)
{
	(void)len;
	return vsprintf(buf, format, args);
}

static tiff_int_t __CDECL gcc_fprintf(FILE *fp, const char *format, ...)
{
	va_list args;
	int ret;
	va_start(args, fmt);
	ret = vfprintf(fp, format, args);
	va_end(args);
	return ret;
}

static tiff_int_t __CDECL gcc_vfprintf(FILE *fp, const char *format, va_list args)
{
	return vfprintf(fp, format, args);
}

static tiff_int_t __CDECL gcc_fputc(tiff_int_t c, FILE *fp)
{
	return fputc((int)c, fp);
}

static tiff_int_t __CDECL gcc_rand(void)
{
	return rand();
}

static void __CDECL gcc_srand(tiff_uint_t seed)
{
	srand((unsigned int)seed);
}

static char *__CDECL gcc_getenv(const char *name)
{
	return getenv(name);
}

static tiff_int_t __CDECL gcc_sigsetjmp(jmp_buf buf, tiff_int_t savesig)
{
	(void)savesig;
	/*
	 * note: mintlib's jmp_buf and Pure-C's have different usages
	 * in the fields, but luckily they are of the same size,
	 * so we can actually do this:
	 */
	return setjmp(buf);
}

static void __CDECL gcc_longjmp(jmp_buf buf, tiff_int_t e)
{
	longjmp(buf, (int)e);
}

#define S(x) tiff_funcs.p_##x = gcc_##x

#else

static int __CDECL get_errno(void)
{
	return errno;
}

static void __CDECL set_errno(int e)
{
	errno = e;
}

#define S(x) tiff_funcs.p_##x = x

#endif


static long __CDECL tiff_zlib_open(void)
{
	SLB *tiff = slb_tiff_get();
	SLB *zlib;
	long ret;

	if (!tiff)
		return -EBADF;
	if (!tiff->handle)
		return -EBADF;
	zlib = slb_zlib_get();
	if (!zlib)
		return -EBADF;
	if (zlib->handle == 0)
	{
		ret = slb_zlib_open(tiff_funcs.slbpath);
		if (ret != 0)
			return ret;
		tiff_funcs.zlib_opened_here = 1;
	}
	tiff_funcs.zlib = zlib;
	return 0;
}


static long __CDECL tiff_jpeg_open(void)
{
	SLB *tiff = slb_tiff_get();
	SLB *jpeg;
	long ret;

	if (!tiff)
		return -EBADF;
	if (!tiff->handle)
		return -EBADF;
	jpeg = slb_jpeglib_get();
	if (jpeg->handle == 0)
	{
		ret = slb_jpeglib_open(tiff_funcs.slbpath);
		if (ret != 0)
			return ret;
		tiff_funcs.jpeg_opened_here = 1;
	}
	tiff_funcs.jpeg = jpeg;
	return 0;
}


static long __CDECL tiff_lzma_open(void)
{
	SLB *tiff = slb_tiff_get();
	SLB *lzma;
	long ret;

	if (!tiff)
		return -EBADF;
	if (!tiff->handle)
		return -EBADF;
	lzma = slb_lzma_get();
	if (lzma->handle == 0)
	{
		ret = slb_lzma_open(tiff_funcs.slbpath);
		if (ret != 0)
			return ret;
		tiff_funcs.lzma_opened_here = 1;
	}
	tiff_funcs.lzma = lzma;
	return 0;
}


long slb_tiff_open(const char *slbpath)
{
	long ret;
	long flags;
	long cpu;
	SLB *tiff = slb_tiff_get();
	
	if (!tiff)
		return -EBADF;
	if (tiff->handle)
		return 0;

	tiff_funcs.struct_size = sizeof(tiff_funcs);
	tiff_funcs.int_size = sizeof(tiff_int_t);
	tiff_funcs.tiff_vernum = TIFF_VERSION;
	tiff_funcs.stderr_location = stderr;

	S(get_errno);
	S(set_errno);
	S(strerror);

	S(memset);
	S(memcpy);
	S(memcmp);
	S(memchr);

	S(strlen);
	S(strcpy);
	S(strcat);
	S(strcmp);
	S(strncmp);

	S(malloc);
	S(calloc);
	S(realloc);
	S(free);

	S(open);
	S(close);
	S(read);
	S(write);
	S(lseek);

	S(fwrite);
	S(sprintf);
	S(snprintf);
	S(vsnprintf);
	S(fprintf);
	S(vfprintf);
	S(fputc);

	S(sigsetjmp);
	S(longjmp);

	S(rand);
	S(srand);
	S(getenv);

#if 0
	/* functions forwarded to jpeglib */
	S(strncpy);
	S(fread);
	S(fflush);
	S(ferror);
	S(exit);
#endif

#undef S

	tiff_funcs.slbpath = slbpath;
	tiff_funcs.zlib_opened_here = 0;
	tiff_funcs.zlib = 0;
	tiff_funcs.p_slb_tiff_zlib_open = tiff_zlib_open;

	tiff_funcs.jpeg_opened_here = 0;
	tiff_funcs.jpeg = 0;
	tiff_funcs.p_slb_tiff_jpeg_open = tiff_jpeg_open;

	tiff_funcs.lzma_opened_here = 0;
	tiff_funcs.lzma = 0;
	tiff_funcs.p_slb_tiff_lzma_open = tiff_lzma_open;

	ret = slb_load(TIFF_SHAREDLIB_NAME, slbpath, TIFF_VERSION, &tiff->handle, &tiff->exec);
	if (ret < 0)
	{
		slb_tiff_close();
		return ret;
	}

	/*
	 * check compile flags; that function should be as simple as to just return a constant
	 * and we can hopefully call it even on mismatched configurations
	 */
	flags = tiff_slb_control(0, 0);
	get_cookie(C__CPU, &cpu);
	if (cpu >= 20)
	{
		/* should be able to use a 000 library, anyways */
	} else
	{
		if (flags & (1L << 16))
		{
			/* cpu is not 020+, but library was compiled for it */
			slb_tiff_close();
			return -EINVAL;
		}
	}
#if defined(__mcoldfire__)
	/* if cpu is cf, but library was not compiled for it... */
	if (!(flags & (1L << 17)))
#else
	/* if cpu is not cf, but library was compiled for it... */
	if (flags & (1L << 17))
#endif
	{
		slb_tiff_close();
		return -EINVAL;
	}
	
	ret = tiff_slb_control(1, &tiff_funcs);
	if (ret < 0)
	{
		slb_tiff_close();
		return ret;
	}
	
	return ret;
}
