/*
 * Copyright (c) 1988-1997 Sam Leffler
 * Copyright (c) 1991-1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _TIFFIOP_
#define	_TIFFIOP_
/*
 * ``Library-private'' definitions.
 */

#include "tif_conf.h"

#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_STRING_H
# include <string.h>
#endif

#if 0
# include <assert.h>
#else
# define assert(x) 
#endif

#if !defined(HAVE_SNPRINTF) && !defined(HAVE__SNPRINTF)
#undef snprintf
#define snprintf _TIFF_snprintf_f
extern int snprintf(char* str, size_t size, const char* format, ...);
#endif

#include "tiffio.h"

#include "tif_dir.h"

#ifndef STRIP_SIZE_DEFAULT
# define STRIP_SIZE_DEFAULT 8192
#endif

#define    streq(a,b)      (strcmp(a,b) == 0)
#define    strneq(a,b,n)   (strncmp(a,b,n) == 0)

#ifndef TRUE
#define	TRUE	1
#define	FALSE	0
#endif

typedef struct client_info {
    struct client_info *next;
    void *data;
    char *name;
} TIFFClientInfoLink;

/*
 * Typedefs for ``method pointers'' used internally.
 * these are deprecated and provided only for backwards compatibility.
 */
typedef unsigned char tidataval_t;    /* internal image data value type */
typedef tidataval_t* tidata_t;        /* reference to internal image data */

typedef void (*TIFFVoidMethod)(TIFF*);
typedef int (*TIFFBoolMethod)(TIFF*);
typedef int (*TIFFPreMethod)(TIFF*, uint16);
typedef int (*TIFFCodeMethod)(TIFF* tif, uint8* buf, tmsize_t size, uint16 sample);
typedef int (*TIFFSeekMethod)(TIFF*, uint32);
typedef void (*TIFFPostMethod)(TIFF* tif, uint8* buf, tmsize_t size);
typedef uint32 (*TIFFStripMethod)(TIFF*, uint32);
typedef void (*TIFFTileMethod)(TIFF*, uint32*, uint32*);

struct tiff {
	char*                tif_name;         /* name of open file */
	int                  tif_fd;           /* open file descriptor */
	int                  tif_mode;         /* open mode (O_*) */
	uint32               tif_flags;
	#define TIFF_FILLORDER   UINT32_C(0x00003) /* natural bit fill order for machine */
	#define TIFF_DIRTYHEADER UINT32_C(0x00004) /* header must be written on close */
	#define TIFF_DIRTYDIRECT UINT32_C(0x00008) /* current directory must be written */
	#define TIFF_BUFFERSETUP UINT32_C(0x00010) /* data buffers setup */
	#define TIFF_CODERSETUP  UINT32_C(0x00020) /* encoder/decoder setup done */
	#define TIFF_BEENWRITING UINT32_C(0x00040) /* written 1+ scanlines to file */
	#define TIFF_SWAB        UINT32_C(0x00080) /* byte swap file information */
	#define TIFF_NOBITREV    UINT32_C(0x00100) /* inhibit bit reversal logic */
	#define TIFF_MYBUFFER    UINT32_C(0x00200) /* my raw data buffer; free on close */
	#define TIFF_ISTILED     UINT32_C(0x00400) /* file is tile, not strip- based */
	#define TIFF_MAPPED      UINT32_C(0x00800) /* file is mapped into memory */
	#define TIFF_POSTENCODE  UINT32_C(0x01000) /* need call to postencode routine */
	#define TIFF_INSUBIFD    UINT32_C(0x02000) /* currently writing a subifd */
	#define TIFF_UPSAMPLED   UINT32_C(0x04000) /* library is doing data up-sampling */
	#define TIFF_STRIPCHOP   UINT32_C(0x08000) /* enable strip chopping support */
	#define TIFF_HEADERONLY  UINT32_C(0x10000) /* read header only, do not process the first directory */
	#define TIFF_NOREADRAW   UINT32_C(0x20000) /* skip reading of raw uncompressed image data */
	#define TIFF_INCUSTOMIFD UINT32_C(0x40000) /* currently writing a custom IFD */
	#define TIFF_BIGTIFF     UINT32_C(0x80000) /* read/write bigtiff */
        #define TIFF_BUF4WRITE  UINT32_C(0x100000) /* rawcc bytes are for writing */
        #define TIFF_DIRTYSTRIP UINT32_C(0x200000) /* stripoffsets/stripbytecount dirty*/
        #define TIFF_PERSAMPLE  UINT32_C(0x400000) /* get/set per sample tags as arrays */
        #define TIFF_BUFFERMMAP UINT32_C(0x800000) /* read buffer (tif_rawdata) points into mmap() memory */
	uint64               tif_diroff;       /* file offset of current directory */
	uint64               tif_nextdiroff;   /* file offset of following directory */
	uint64*              tif_dirlist;      /* list of offsets to already seen directories to prevent IFD looping */
	uint16               tif_dirlistsize;  /* number of entries in offset list */
	uint16               tif_dirnumber;    /* number of already seen directories */
	TIFFDirectory        tif_dir;          /* internal rep of current directory */
	TIFFDirectory        tif_customdir;    /* custom IFDs are separated from the main ones */
	union {
		TIFFHeaderCommon common;
		TIFFHeaderClassic classic;
		TIFFHeaderBig big;
	} tif_header;
	uint16               tif_header_size;  /* file's header block and its length */
	uint32               tif_row;          /* current scanline */
	uint16               tif_curdir;       /* current directory (index) */
	uint32               tif_curstrip;     /* current strip for read/write */
	uint64               tif_curoff;       /* current offset for read/write */
	uint64               tif_dataoff;      /* current offset for writing dir */
	/* SubIFD support */
	uint16               tif_nsubifd;      /* remaining subifds to write */
	uint64               tif_subifdoff;    /* offset for patching SubIFD link */
	/* tiling support */
	uint32               tif_col;          /* current column (offset by row too) */
	uint32               tif_curtile;      /* current tile for read/write */
	tmsize_t             tif_tilesize;     /* # of bytes in a tile */
	/* compression scheme hooks */
	int                  tif_decodestatus;
	TIFFBoolMethod       tif_fixuptags;    /* called in TIFFReadDirectory */
	TIFFBoolMethod       tif_setupdecode;  /* called once before predecode */
	TIFFPreMethod        tif_predecode;    /* pre- row/strip/tile decoding */
	TIFFBoolMethod       tif_setupencode;  /* called once before preencode */
	int                  tif_encodestatus;
	TIFFPreMethod        tif_preencode;    /* pre- row/strip/tile encoding */
	TIFFBoolMethod       tif_postencode;   /* post- row/strip/tile encoding */
	TIFFCodeMethod       tif_decoderow;    /* scanline decoding routine */
	TIFFCodeMethod       tif_encoderow;    /* scanline encoding routine */
	TIFFCodeMethod       tif_decodestrip;  /* strip decoding routine */
	TIFFCodeMethod       tif_encodestrip;  /* strip encoding routine */
	TIFFCodeMethod       tif_decodetile;   /* tile decoding routine */
	TIFFCodeMethod       tif_encodetile;   /* tile encoding routine */
	TIFFVoidMethod       tif_close;        /* cleanup-on-close routine */
	TIFFSeekMethod       tif_seek;         /* position within a strip routine */
	TIFFVoidMethod       tif_cleanup;      /* cleanup state routine */
	TIFFStripMethod      tif_defstripsize; /* calculate/constrain strip size */
	TIFFTileMethod       tif_deftilesize;  /* calculate/constrain tile size */
	uint8*               tif_data;         /* compression scheme private data */
	/* input/output buffering */
	tmsize_t             tif_scanlinesize; /* # of bytes in a scanline */
	tmsize_t             tif_scanlineskew; /* scanline skew for reading strips */
	uint8*               tif_rawdata;      /* raw data buffer */
	tmsize_t             tif_rawdatasize;  /* # of bytes in raw data buffer */
	tmsize_t             tif_rawdataoff;   /* rawdata offset within strip */
	tmsize_t             tif_rawdataloaded;/* amount of data in rawdata */
	uint8*               tif_rawcp;        /* current spot in raw buffer */
	tmsize_t             tif_rawcc;        /* bytes unread from raw buffer */
	/* memory-mapped file support */
	uint8*               tif_base;         /* base of mapped file */
	tmsize_t             tif_size;         /* size of mapped file region (bytes, thus tmsize_t) */
	TIFFMapFileProc      tif_mapproc;      /* map file method */
	TIFFUnmapFileProc    tif_unmapproc;    /* unmap file method */
	/* input/output callback methods */
	thandle_t            tif_clientdata;   /* callback parameter */
	TIFFReadWriteProc    tif_readproc;     /* read method */
	TIFFReadWriteProc    tif_writeproc;    /* write method */
	TIFFSeekProc         tif_seekproc;     /* lseek method */
	TIFFCloseProc        tif_closeproc;    /* close method */
	TIFFSizeProc         tif_sizeproc;     /* filesize method */
	/* post-decoding support */
	TIFFPostMethod       tif_postdecode;   /* post decoding routine */
	/* tag support */
	TIFFField**          tif_fields;       /* sorted table of registered tags */
	size_t               tif_nfields;      /* # entries in registered tag table */
	const TIFFField*     tif_foundfield;   /* cached pointer to already found tag */
	TIFFTagMethods       tif_tagmethods;   /* tag get/set/print routines */
	TIFFClientInfoLink*  tif_clientinfo;   /* extra client information. */
	/* Backward compatibility stuff. We need these two fields for
	 * setting up an old tag extension scheme. */
	TIFFFieldArray*      tif_fieldscompat;
	size_t               tif_nfieldscompat;
};

#define isPseudoTag(t) (t > 0xffff)            /* is tag value normal or pseudo */

#define isTiled(tif) (((tif)->tif_flags & TIFF_ISTILED) != 0)
#define isMapped(tif) (((tif)->tif_flags & TIFF_MAPPED) != 0)
#define isFillOrder(tif, o) (((tif)->tif_flags & (o)) != 0)
#define isUpSampled(tif) (((tif)->tif_flags & TIFF_UPSAMPLED) != 0)
#define TIFFReadFile(tif, buf, size) \
	((*(tif)->tif_readproc)((tif)->tif_clientdata,(buf),(size)))
#define TIFFWriteFile(tif, buf, size) \
	((*(tif)->tif_writeproc)((tif)->tif_clientdata,(buf),(size)))
#define TIFFSeekFile(tif, off, whence) \
	((tif)->tif_seekproc?((*(tif)->tif_seekproc)((tif)->tif_clientdata,off,whence)):ll_uint64_zero)
#define TIFFCloseFile(tif) \
	((*(tif)->tif_closeproc)((tif)->tif_clientdata))
#define TIFFGetFileSize(tif) \
	((*(tif)->tif_sizeproc)((tif)->tif_clientdata))
#define TIFFMapFileContents(tif, paddr, psize) \
	((*(tif)->tif_mapproc)((tif)->tif_clientdata,(paddr),(psize)))
#define TIFFUnmapFileContents(tif, addr, size) \
	((*(tif)->tif_unmapproc)((tif)->tif_clientdata,(addr),(size)))

/*
 * Default Read/Seek/Write definitions.
 */
#ifndef ReadOK
#define ReadOK(tif, buf, size) (TIFFReadFile((tif),(buf),(size))==(size))
#endif
#ifndef SeekOK
#define SeekOK(tif, off) _TIFFSeekOK(tif, off)
#endif
#ifndef WriteOK
#define WriteOK(tif, buf, size) (TIFFWriteFile((tif),(buf),(size))==(size))
#endif

/* NB: the uint32 casts are to silence certain ANSI-C compilers */
#define TIFFhowmany_32(x, y) (((uint32)(x) < (UINT32_C(0xffffffff) - (uint32)((y)-1))) ? \
			   ((((uint32)(x))+(((uint32)(y))-1))/((uint32)(y))) : \
			   0U)
/* Variant of TIFFhowmany_32() that doesn't return 0 if x close to MAXUINT. */
/* Caution: TIFFhowmany_32_maxuint_compat(x,y)*y might overflow */
#define TIFFhowmany_32_maxuint_compat(x, y) \
			   (((uint32)(x) / (uint32)(y)) + ((((uint32)(x) % (uint32)(y)) != 0) ? 1 : 0))
#define TIFFhowmany8_32(x) (((x)&0x07)?((uint32)(x)>>3)+1:(uint32)(x)>>3)
#define TIFFroundup_32(x, y) (TIFFhowmany_32(x,y)*(y))
#define TIFFhowmany_64(x, y) ll_uint64_div(ll_uint64_sub(ll_uint64_add(x, y), ll_uint64_make(0, 1)), (y))
#define TIFFhowmany8_64(x) ((ll_uint64_to_uint32(x) & 0x07) ? ll_uint64_add(ll_uint64_rshift(x, 3), ll_uint64_make(0, 1)) : ll_uint64_rshift(x, 3))
#define TIFFroundup_64(x, y) ll_uint64_mul(TIFFhowmany_64(x,y),(y))

/* Safe multiply which returns zero if there is an integer overflow */
#define TIFFSafeMultiply(t,v,m) ((((t)(m) != (t)0) && (((t)(((v)*(m))/(m))) == (t)(v))) ? (t)((v)*(m)) : (t)0)
#define TIFFSafeMultiply64(t,v,m) ((!ll_uint64_iszero(m) && ll_uint64_eq(ll_uint64_div(ll_uint64_mul((v), (m)), (m)), (v))) ? ll_uint64_mul((v), (m)) : ll_uint64_zero)

#define TIFFmax(A,B) ((A)>(B)?(A):(B))
#define TIFFmin(A,B) ((A)<(B)?(A):(B))

#define TIFFArrayCount(a) (sizeof (a) / sizeof ((a)[0]))

/*
  Support for large files.

  Windows read/write APIs support only 'unsigned int' rather than 'size_t'.
  Windows off_t is only 32-bit, even in 64-bit builds.
*/
/*
  Use fseeko() and ftello() if they are available since they use
  'off_t' rather than 'long'.  It is wrong to use fseeko() and
  ftello() only on systems with special LFS support since some systems
  (e.g. FreeBSD) support a 64-bit off_t by default.

  For MinGW, __MSVCRT_VERSION__ must be at least 0x800 to expose these
  interfaces. The MinGW compiler must support the requested version.  MinGW
  does not distribute the CRT (it is supplied by Microsoft) so the correct CRT
  must be available on the target computer in order for the program to run.
*/
#if defined(HAVE_FSEEKO) && !(defined(__atarist__) || defined(__TOS__))
#  define fseek(stream,offset,whence)  fseeko(stream,offset,whence)
#  define ftell(stream,offset,whence)  ftello(stream,offset,whence)
#endif
#if defined(__WIN32__) && \
        !(defined(_MSC_VER) && _MSC_VER < 1400) && \
        !(defined(__MSVCRT_VERSION__) && __MSVCRT_VERSION__ < 0x800)
typedef unsigned int TIFFIOSize_t;
#define _TIFF_lseek_f(fildes,offset,whence)  _lseeki64(fildes,/* __int64 */ offset,whence)
/* #define _TIFF_tell_f(fildes) /\* __int64 *\/ _telli64(fildes) */
#define _TIFF_fseek_f(stream,offset,whence) _fseeki64(stream,/* __int64 */ offset,whence)
#define _TIFF_fstat_f(fildes,stat_buff) _fstati64(fildes,/* struct _stati64 */ stat_buff)
/* #define _TIFF_ftell_f(stream) /\* __int64 *\/ _ftelli64(stream) */
/* #define _TIFF_stat_f(path,stat_buff) _stati64(path,/\* struct _stati64 *\/ stat_buff) */
#define _TIFF_stat_s struct _stati64
#define _TIFF_off_t __int64
#else
typedef size_t TIFFIOSize_t;
#define _TIFF_lseek_f(fildes,offset,whence) lseek(fildes,offset,whence)
/* #define _TIFF_tell_f(fildes) (_TIFF_lseek_f(fildes,0,SEEK_CUR)) */
#define _TIFF_fseek_f(stream,offset,whence) fseek(stream,offset,whence)
#define _TIFF_fstat_f(fildes,stat_buff) fstat(fildes,stat_buff)
/* #define _TIFF_ftell_f(stream) ftell(stream) */
/* #define _TIFF_stat_f(path,stat_buff) stat(path,stat_buff) */
#define _TIFF_stat_s struct stat
#define _TIFF_off_t off_t
#if LL_HAVE_UINT64
#define ll_uint64_to_off_t(x) ((off_t)(x))
#else
#define ll_uint64_to_off_t(x) ((off_t)ll_uint64_to_size_t(x))
#endif
#endif

#if defined(__has_attribute) && defined(__clang__)
#if __has_attribute(no_sanitize)
#define TIFF_NOSANITIZE_UNSIGNED_INT_OVERFLOW __attribute__((no_sanitize("unsigned-integer-overflow")))
#else
#define TIFF_NOSANITIZE_UNSIGNED_INT_OVERFLOW
#endif
#else
#define TIFF_NOSANITIZE_UNSIGNED_INT_OVERFLOW
#endif


#if defined(__cplusplus)
extern "C" {
#endif
int _TIFFgetMode(const char* mode, const char* module);
int _TIFFNoRowEncode(TIFF* tif, uint8* pp, tmsize_t cc, uint16 s);
int _TIFFNoStripEncode(TIFF* tif, uint8* pp, tmsize_t cc, uint16 s);
int _TIFFNoTileEncode(TIFF*, uint8* pp, tmsize_t cc, uint16 s);
int _TIFFNoRowDecode(TIFF* tif, uint8* pp, tmsize_t cc, uint16 s);
int _TIFFNoStripDecode(TIFF* tif, uint8* pp, tmsize_t cc, uint16 s);
int _TIFFNoTileDecode(TIFF*, uint8* pp, tmsize_t cc, uint16 s);
void _TIFFNoPostDecode(TIFF* tif, uint8* buf, tmsize_t cc);
int _TIFFNoPreCode(TIFF* tif, uint16 s);
int _TIFFNoSeek(TIFF* tif, uint32 off);
void _TIFFSwab16BitData(TIFF* tif, uint8* buf, tmsize_t cc);
void _TIFFSwab24BitData(TIFF* tif, uint8* buf, tmsize_t cc);
void _TIFFSwab32BitData(TIFF* tif, uint8* buf, tmsize_t cc);
void _TIFFSwab64BitData(TIFF* tif, uint8* buf, tmsize_t cc);
int TIFFFlushData1(TIFF* tif);
int TIFFDefaultDirectory(TIFF* tif);
void _TIFFSetDefaultCompressionState(TIFF* tif);
tiff_int_t TIFF_API _TIFFRewriteField(TIFF *, ttag_t, tiff_enum(TIFFDataType), tmsize_t, void *);
tiff_int_t TIFF_API TIFFSetCompressionScheme(TIFF* tif, tiff_int_t scheme);
int TIFFSetDefaultCompressionState(TIFF* tif);
uint32 _TIFFDefaultStripSize(TIFF* tif, uint32 s);
void _TIFFDefaultTileSize(TIFF* tif, uint32* tw, uint32* th);
int _TIFFDataSize(TIFFDataType type);

void _TIFFsetByteArray(void**, void*, uint32);
void _TIFFsetString(char**, char*);
void _TIFFsetShortArray(uint16**, uint16*, uint32);
void _TIFFsetLongArray(uint32**, uint32*, uint32);
void _TIFFsetFloatArray(float**, float*, uint32);
void _TIFFsetDoubleArray(double**, double*, uint32);

void _TIFFprintAscii(FILE*, const char*);
void _TIFFprintAsciiTag(FILE*, const char*, const char*);

extern TIFFErrorHandler _TIFFwarningHandler;
extern TIFFErrorHandler _TIFFerrorHandler;
extern TIFFErrorHandlerExt _TIFFwarningHandlerExt;
extern TIFFErrorHandlerExt _TIFFerrorHandlerExt;

uint32 TIFF_API _TIFFMultiply32(TIFF*, uint32, uint32, const char*);
uint64 TIFF_API _TIFFMultiply64(TIFF*, uint64, uint64, const char*);
void* TIFF_API _TIFFCheckMalloc(TIFF*, tmsize_t, tmsize_t, const char*);
void* TIFF_API _TIFFCheckRealloc(TIFF*, void*, tmsize_t, tmsize_t, const char*);

double _TIFFUInt64ToDouble(uint64);
float _TIFFUInt64ToFloat(uint64);

tmsize_t
_TIFFReadEncodedStripAndAllocBuffer(TIFF* tif, uint32 strip,
                                    void **buf, tmsize_t bufsizetoalloc,
                                    tmsize_t size_to_read);
tmsize_t
_TIFFReadEncodedTileAndAllocBuffer(TIFF* tif, uint32 tile,
                                    void **buf, tmsize_t bufsizetoalloc,
                                    tmsize_t size_to_read);
tmsize_t
_TIFFReadTileAndAllocBuffer(TIFF* tif,
                            void **buf, tmsize_t bufsizetoalloc,
                            uint32 x, uint32 y, uint32 z, uint16 s);
int _TIFFSeekOK(TIFF* tif, toff_t off);

tiff_int_t TIFFInitDumpMode(TIFF*, tiff_uint_t);
#ifdef PACKBITS_SUPPORT
tiff_int_t TIFFInitPackBits(TIFF*, tiff_uint_t);
#endif
#ifdef CCITT_SUPPORT
tiff_int_t TIFFInitCCITTRLE(TIFF *, tiff_uint_t);
tiff_int_t TIFFInitCCITTRLEW(TIFF*, tiff_uint_t);
tiff_int_t TIFFInitCCITTFax3(TIFF*, tiff_uint_t);
tiff_int_t TIFFInitCCITTFax4(TIFF*, tiff_uint_t);
#endif
#ifdef THUNDER_SUPPORT
tiff_int_t TIFFInitThunderScan(TIFF*, tiff_uint_t);
#endif
#ifdef NEXT_SUPPORT
tiff_int_t TIFFInitNeXT(TIFF*, tiff_uint_t);
#endif
#ifdef LZW_SUPPORT
tiff_int_t TIFFInitLZW(TIFF*, tiff_uint_t);
#endif
#ifdef OJPEG_SUPPORT
tiff_int_t TIFFInitOJPEG(TIFF*, tiff_uint_t);
#endif
#ifdef JPEG_SUPPORT
tiff_int_t TIFFInitJPEG(TIFF*, tiff_uint_t);
int TIFFJPEGIsFullStripRequired(TIFF*);
#endif
#ifdef JBIG_SUPPORT
tiff_int_t TIFFInitJBIG(TIFF*, tiff_uint_t);
#endif
#ifdef ZIP_SUPPORT
tiff_int_t TIFFInitZIP(TIFF*, tiff_uint_t);
#endif
#ifdef PIXARLOG_SUPPORT
tiff_int_t TIFFInitPixarLog(TIFF*, tiff_uint_t);
#endif
#ifdef LOGLUV_SUPPORT
tiff_int_t TIFFInitSGILog(TIFF*, tiff_uint_t);
#endif
#ifdef LZMA_SUPPORT
tiff_int_t TIFFInitLZMA(TIFF*, tiff_uint_t);
#endif
#ifdef ZSTD_SUPPORT
tiff_int_t TIFFInitZSTD(TIFF*, tiff_uint_t);
#endif
#ifdef WEBP_SUPPORT
tiff_int_t TIFFInitWebP(TIFF*, tiff_uint_t);
#endif
extern const TIFFCodec _TIFFBuiltinCODECS[];

int TIFFFillStrip(TIFF *tif, uint32 strip);
int TIFFFillTile(TIFF *tif, uint32 tile);
int TIFFReInitJPEG_12(TIFF * tif, int scheme, int is_encode);
int TIFFJPEGIsFullStripRequired_12(TIFF * tif);

#if defined(__cplusplus)
}
#endif

#if defined(TIFF_SLB) && !defined(TIFF_TOOL_HACKS) && (defined(__atarist__) || defined(__TOS__))
#include "libslb.h"
#endif

#endif /* _TIFFIOP_ */
